# -*- coding: utf-8 -*-
""" 
@author: catherine wei
@contact: EMAIL@contact: catherine@oddmeta.com
@software: PyCharm 
@file: odd_log.py 
@info: 消息模版
"""

import logging
from logging import handlers
import oddasr.odd_asr_config as config


def _logging():
    FORMAT = "%(asctime)s %(levelname)s %(filename)s:%(lineno)s (%(process)s-%(thread)s) - %(message)s "
    DATE = '%Y-%m-%d %H:%M:%S'

    format = logging.Formatter(FORMAT, DATE)

    import os
    if not os.path.exists(config.log_path):
        os.makedirs(config.log_path)

    logfile = config.log_path + config.log_file
    
    log = logging.getLogger(logfile)

    th = handlers.TimedRotatingFileHandler(filename=logfile, when='MIDNIGHT', backupCount=10, encoding='utf-8')
    th.setFormatter(format)
    log.addHandler(th)

    stdout = logging.StreamHandler()
    stdout.setFormatter(format)
    log.addHandler(stdout)

    if config.Debug:
        enableProtoPrint = False
        if enableProtoPrint:
            logging.basicConfig(level=logging.DEBUG, format=FORMAT, datefmt=DATE)
        else:
            ch = logging.StreamHandler()
            ch.setFormatter(format)
            log.addHandler(ch)

    log.setLevel(config.log_level)
    
    return log


logger = _logging()






# -*- coding: utf-8 -*-
""" 
@author: catherine wei
@contact: EMAIL@contact: catherine@oddmeta.com
@software: PyCharm 
@file: testStreamAPI.py 
@info: 消息模版
"""

"""Client example using the asyncio API."""

'''
    client --> server: connect
    client --> server: TCmdApppyAsrReq, msg_type = StartTranscription;
    server --> client: TCmdApplyAsrRes, msg_type = MSG_SUBSCRIBE_INIT_RES;
    client --> server: TCMDTranscribeReq, msg_type = MSG_TRANSCRIBE_REQ;
    server --> client: TCMDTranscribeRes, msg_type = MSG_TRANSCRIBE_RES;

    TCMDTranscribeReq

'''

import asyncio
import json
import time
import websockets
import enum

from websockets.asyncio.client import connect

from oddasr.log import logger
import oddasr.odd_asr_config as config
import oddasr.logic.proto as proto

class odd_asr_state(enum.Enum):
    EM_ASR_STATE_IDLE = 0,
    EM_ASR_STATE_APPLYING = 1,
    EM_ASR_STATE_APLLY_OK = 2,
    EM_ASR_STATE_APPLY_FAILED = 3,
    EM_ASR_STATE_RECOGNIZING = 4,

# state: odd_asr_state = odd_asr_state.EM_ASR_STATE_IDLE

class OddWsClient:
    def __init__(self, uri):
        self.uri = uri
        self.websocket = None
        self.state = odd_asr_state.EM_ASR_STATE_IDLE
        self.timestamp_apply = 0
        self.timestamp_first_packet = 0
        self.timestamp_first_text = 0
        self.timestamp_last_packet = 0
        self.timestamp_last_text = 0
        self.timestamp_sentence_begin = 0
        self.timestamp_sentence_end = 0

ws_client_set = set()


async def receive_messages(websocket):
    '''
    receive message from server
    :param websocket: WebSocket connection object
    '''

    while True:
        try:
            # 接收服务端消息
            message = await websocket.recv()
            # logger.debug(f"<<< {message}")

            odd_ws_client: OddWsClient = None
            for client in ws_client_set:
                if client.websocket == websocket:
                    odd_ws_client = client

            match odd_ws_client.state:
                case odd_asr_state.EM_ASR_STATE_IDLE:
                    continue
                case odd_asr_state.EM_ASR_STATE_APPLYING:
                    response = json.loads(message)
                    logger.debug(f"<<< {response}")
                    res = proto.TOddAsrApplyRes()
                    res = proto.obj_from_dict_recursive(res, response)
                    if res.header.name == "SentenceBegin":
                        if res.header.status == 0:
                            odd_ws_client.state = odd_asr_state.EM_ASR_STATE_APLLY_OK
                            logger.info("client doInit ok")
                            odd_ws_client.state = odd_asr_state.EM_ASR_STATE_RECOGNIZING
                            odd_ws_client.timestamp_first_text = time.time()
                            odd_ws_client.timestamp_sentence_begin = res.payload.time
                        else:
                            odd_ws_client.state = odd_asr_state.EM_ASR_STATE_APPLY_FAILED
                    odd_ws_client.timestamp_apply = time.time()
                    continue
                case odd_asr_state.EM_ASR_STATE_APLLY_OK:
                    continue
                case odd_asr_state.EM_ASR_STATE_APPLY_FAILED:
                    continue
                case odd_asr_state.EM_ASR_STATE_RECOGNIZING:
                    response = json.loads(message)
                    logger.debug(f"<<< {response}")
                    res = proto.TOddAsrTranscribeRes()
                    res = proto.obj_from_dict_recursive(res, response)
                    if res.header.name == "SentenceEnd":
                        odd_ws_client.timestamp_sentence_end = time.time()
                    elif res.header.name == "SentenceBegin":
                        odd_ws_client.timestamp_sentence_begin = res.payload.time
                    elif res.header.name == "TranscriptionResultChanged":
                        odd_ws_client.timestamp_last_packet = time.time()
                    elif res.header.name == "TranscriptionCompleted":
                        odd_ws_client.timestamp_last_text = time.time()
                    else:
                        logger.error(f"unknown message type={res.header.name}")
                case _:
                    continue

        except websockets.exceptions.ConnectionClosedOK:
            logger.error("Connection closed gracefully")
            break
        except Exception as e:
            logger.error(f"Receive error: {e}")
            break

async def send_messages(websocket, file):
    offset = 0
    total_length = 0
    chunk_size = 9600*2

    odd_ws_client: OddWsClient = None
    for client in ws_client_set:
        if client.websocket == websocket:
            odd_ws_client = client

    with open(file, "rb") as f:
        data = f.read()

    total_length = len(data)
    logger.debug(f"total_length={total_length}")

    while True:
        await asyncio.sleep(0.1)
        match odd_ws_client.state:
            case odd_asr_state.EM_ASR_STATE_IDLE:
                req = proto.TOddAsrApplyReq()
                req = {"name": "StartTranscription", "message_id": "", "token": "", "task_id": ""}

                str = proto.obj_to_dict_recursive(req)
                logger.debug(f">>>client doInit req: {str}")

                await websocket.send(json.dumps(req))
                logger.debug(f">>>client doInit req sent, timestamp_apply={odd_ws_client.timestamp_apply}")
                odd_ws_client.state = odd_asr_state.EM_ASR_STATE_APPLYING
                continue
            case odd_asr_state.EM_ASR_STATE_APPLYING:
                continue
            case odd_asr_state.EM_ASR_STATE_APLLY_OK:
                continue
            case odd_asr_state.EM_ASR_STATE_APPLY_FAILED:
                continue
            case odd_asr_state.EM_ASR_STATE_RECOGNIZING:

                chunk = data[offset:offset + chunk_size]
                await websocket.send(chunk)

                if odd_ws_client.timestamp_first_packet == 0:
                    odd_ws_client.timestamp_first_packet = time.time()
                    logger.debug(f"send chunk {offset}, odd_ws_client.timestamp_first_packet={odd_ws_client.timestamp_first_packet}")
                else:
                    logger.debug(f"send chunk {offset}")

                offset += chunk_size

                if offset >= total_length:
                    logger.info("send end")
                    break
                await asyncio.sleep(0.2)
                continue
            case _:
                continue

async def hello(server_url, file):
    client = OddWsClient(server_url)
    ws_client_set.add(client)

    async with connect(server_url) as websocket:
        client.websocket = websocket
        send_task = asyncio.create_task(send_messages(websocket, file))
        receive_task = asyncio.create_task(receive_messages(websocket))

        # 等待接收任务完成
        await asyncio.gather(send_task, receive_task)
        
if __name__ == "__main__":
    import argparse
    import os

    parser = argparse.ArgumentParser(description="Your WAV file need to recognize to text.")
    parser.add_argument("audio_path", type=str, help="file path of your input WAV.")
    # Add server address parameter, allow it to be empty, set default value to None
    parser.add_argument("--server_url", type=str, default=None, help="Server WebSocket URL")
    # Add concurrent connection number parameter, allow it to be empty, set default value to None
    parser.add_argument("--concurrency", type=int, default=None, help="Number of concurrent connections")
    args = parser.parse_args()

    file = args.audio_path
    default_server_url = "ws://"+ config.WS_HOST +":"+ str(config.WS_PORT) +"/v1/asr"
    
    server_url = args.server_url if args.server_url is not None else default_server_url
    concurrency = args.concurrency if args.concurrency is not None else 1

    print(f"Current working directory: {os.getcwd()}")
    print(f"Full file path: {os.path.abspath(file)}")
    print(f"ASR server url: {server_url}, concurrency={concurrency}")
    

    if not os.path.exists(file):
        print(f"File not found: {file}")
        exit(1) 

    # detect current test file is wav file
    if not file.endswith(".wav") and not file.endswith(".pcm"):
        print(f"File format error: {file}. Please input wav or pcm file.")
        exit(1)

    # check file format, sample rate must be 16000, sample width must be 16bit, channels must be 1
    if file.endswith(".wav"):
        import soundfile as sf
        with sf.SoundFile(file) as f:
            if f.samplerate != 16000:
                print(f"File sample rate error: {file}. Please input 16000 sample rate, while input {f.samplerate}")
                exit(1)
            if f.subtype != 'PCM_16':
                print(f"File sample width error: {file}. Please input 16bit sample width, while input {f.subtype}")
                exit(1)
            if f.channels != 1:
                print(f"File channels error: {file}. Please input 1 channel, while input {f.channels}")
                exit(1)

    async def run_concurrent_tests():
        tasks = [hello(server_url, file) for _ in range(concurrency)]
        await asyncio.gather(*tasks)

    asyncio.run(run_concurrent_tests())



import time
import numpy as np
import subprocess as sp
import threading
from funasr import AutoModel


# Wakeup words
ODD_WAKEUP_WORD = ["你好小落,小落小落,小落同学"]

# Conference commands
ODD_CONF_COMMANDS = [
    "开始会议",
    "结束会议",
    "邀请参与者",
    "取消邀请",
    "静音参与者",
    "取消静音",
    "发送双流",
    "停止双流",
    "切换摄像头",
    "切换麦克风",
    "打开录音",
    "关闭录音",
]

# IOT commands
ODD_IOT_COMMANDS = [
    "打开灯光",
    "关闭灯光",
    "打开空调",
    "关闭空调",
    "打开窗帘",
    "关闭窗帘",
    "打开电视",
    "关闭电视",
    "打开终端",
    "关闭终端",
]

# Music control commands
ODD_MUSIC_COMMANDS = [
    "播放音乐",
    "增大音量",
    "减小音量",
    "继续播放",
    "暂停播放",
    "上一首",
    "下一首",
    "单曲循环",
    "随机模式",
    "列表循环",
]

# Map control commands
ODD_MAP_COMMANDS = [
    "取消导航",
    "退出导航",
    "放大地图",
    "查看全程",
    "缩小地图",
    "不走高速",
    "躲避拥堵",
    "避免收费",
    "高速优先",
]

# General commands
ODD_GENERAL_COMMANDS = [
    "上一页",
    "下一页",
    "上一个",
    "下一个",
    "换一批", 
    "返回",
    "退出",
    "帮助",
    "重启",
    "关机",
    "重启系统",
    "关机系统",
    "再见",
    "拜拜",
    "谢谢",
    "谢谢小落",
]

ODD_VOICE_COMMANDS = ODD_CONF_COMMANDS + ODD_IOT_COMMANDS + ODD_MUSIC_COMMANDS + ODD_MAP_COMMANDS + ODD_GENERAL_COMMANDS

# config
odd_voice_assist_cfg = {
    "commands": {
        "wakeup_word": ODD_WAKEUP_WORD,
        "voice_commands": ODD_VOICE_COMMANDS,
    },

    "audio_fmt": {
        "chunk_size_ms": 200,
        "stride_ms": 480,
        "sample_rate": 16000,
        "format": np.int16,
        "channels": 1,
        "audio_queue_len": 10, # 有检测到VAD后，会将chunk加入到队列中，此参数代表最大队列长度
    }
}


class OddVoiceAssistant:
    """
    Voice Assistant class
    """
    running: bool = False
    is_speaking: bool = False
    capture_thread: threading.Thread = None
    ffmpeg_proc = None
    ffmpeg_proc_event: threading.Event = None
    data_lock: threading.Lock = None

    def __init__(self, wakeup_word=["你好小落,小落小落,小落同学"]):
        
        self.ffmpeg_proc = None
        self.ffmpeg_proc_event = threading.Event()
        self._init_audio_buffers()
        self.data_lock = threading.Lock()

        wakeup_word = ','.join(ODD_VOICE_COMMANDS) if isinstance(wakeup_word, list) else wakeup_word
        self._load_model(wakeup_word)

    def _load_model(self, wakeup_word):
        """load models"""
        try:
            # init vad model
            self.vad_model = AutoModel(
                model="fsmn-vad",
                model_revision="v2.0.4",
                disable_pbar=True
            )
            
            # init kws model
            self.kws_model = AutoModel(
                model="iic/speech_charctc_kws_phone-xiaoyun",
                keywords=wakeup_word,
                output_dir="./outputs/debug",
                device='cpu',
                disable_pbar=True
            )
            
            # cache for vad model and kws_model
            self.vad_cache = {}
            self.kws_cache = {}
            
            # chunk size for vad model and kws model
            self.vad_chunk_samples = int(odd_voice_assist_cfg["audio_fmt"]["chunk_size_ms"] * odd_voice_assist_cfg["audio_fmt"]["sample_rate"] / 1000)
            self.kws_chunk_stride = int(odd_voice_assist_cfg["audio_fmt"]["stride_ms"] * odd_voice_assist_cfg["audio_fmt"]["sample_rate"] / 1000)
            
        except Exception as e:
            raise RuntimeError(f"init models failed: {str(e)}")


    def _init_audio_buffers(self):
        """init audio buffers"""
        self.raw_buffer = np.array([], dtype=np.float32)  # raw audio buffer
        self.wakeup_queue = []                            # wakeup model process queue
        self.MAX_QUEUE_SIZE = odd_voice_assist_cfg["audio_fmt"]["audio_queue_len"]

    def start(self):
        """start odd voice assistant"""
        # start audio capture thread
        self.capture_thread = threading.Thread(target=self._audio_capture_thread_wrapper, daemon=True)
        self.capture_thread.start()
        
        # wait for audio capture initialization
        if not self.ffmpeg_proc_event.wait(timeout=5):
            print("audio capture initialization timeout")
            return False
        
        # start process thread
        self.process_thread = threading.Thread(target=self._audio_process_thread_wrapper, daemon=True)
        self.process_thread.start()
        
        return True

    def _audio_capture_thread_wrapper(self):
        """audio capture thread"""
        try:
            FFMPEG_PATH = "ffmpeg"
            cmd = [
                FFMPEG_PATH,
                '-loglevel', '0',
                '-ar', str(odd_voice_assist_cfg["audio_fmt"]["sample_rate"]),
                '-ac', str(odd_voice_assist_cfg["audio_fmt"]["channels"]),
                '-f', 'pulse',
                '-i', 'default',
                '-fflags', 'nobuffer',
                '-flags', 'low_delay',
                '-f', 's16le',
                '-'
            ]
            
            self.ffmpeg_proc = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.DEVNULL, bufsize=1024*64)
            self.ffmpeg_proc_event.set()
            
            while self.running and self.ffmpeg_proc.poll() is None:
                time.sleep(0.1)
                
        except Exception as e:
            print(f"exception in audio capture thread: {str(e)}")
            self.running = False
        finally:
            self.ffmpeg_proc_event.set()

    def _audio_process_thread_wrapper(self):
        """audio process thread"""
        try:
            while self.running:
                # capture audio
                pcm_data = self.ffmpeg_proc.stdout.read(self.vad_chunk_samples * 2)
                if not pcm_data:
                    break
                
                # convert pcm to np array
                audio_array = np.frombuffer(pcm_data, dtype=odd_voice_assist_cfg["audio_fmt"]["format"])
                audio_array = audio_array.astype(np.float32) / np.iinfo(odd_voice_assist_cfg["audio_fmt"]["format"]).max

                # update audio buffers
                with self.data_lock:
                    self.raw_buffer = np.concatenate((self.raw_buffer, audio_array))

                chunk = None
                
                while len(self.raw_buffer) >= self.vad_chunk_samples:
                    with self.data_lock:
                        chunk = self.raw_buffer[:self.vad_chunk_samples]
                        self.raw_buffer = self.raw_buffer[self.vad_chunk_samples:]

                    if chunk is None:
                        continue

                    # run VAD detection
                    self._vad_detection(chunk)
                    
                    # process wakeup model if vad detected
                    if self.is_speaking:
                        self._process_wakeup(chunk)
                        
        except Exception as e:
            print(f"exception in audio process thread: {str(e)}")
        finally:
            self._cleanup()

    def _vad_detection(self, chunk):
        """run VAD detection"""
        vad_result = self.vad_model.generate(
            input=chunk,
            cache=self.vad_cache,
            is_final=False,
            chunk_size=odd_voice_assist_cfg["audio_fmt"]["chunk_size_ms"]
        )
        
        if vad_result and len(vad_result[0]["value"]) > 0:
            for value in vad_result[0]['value']:
                if value[1] == -1:  # VAD start
                    self.is_speaking = True
                    print(f"vad deteccted at: {value[0]}ms")
                else:  # VAD end
                    self.is_speaking = False
                    print(f"vad end: {value[1]}ms")

    def _process_wakeup(self, chunk):
        """process wakeup model"""
        self.wakeup_queue.append(chunk)
        
        # if queue size > MAX_QUEUE_SIZE, pop and drop the first chunk
        if len(self.wakeup_queue) > self.MAX_QUEUE_SIZE:
            self.wakeup_queue.pop(0)
            
        # execute kws model
        if len(self.wakeup_queue) >= 3:  # 至少3个chunk
            input_data = np.concatenate(self.wakeup_queue)
            res = self.kws_model.generate(
                input=input_data,
                cache=self.kws_cache,
                chunk_size=odd_voice_assist_cfg["audio_fmt"]["stride_ms"]
            )
            print(f"kws result: {res}")

            if res and res[0]['text'] != 'rejected':
                print(f"wakeup success: {res[0]}")
                self.wakeup_queue.clear()

    def _cleanup(self):
        """cleanup resources"""
        if self.ffmpeg_proc and self.ffmpeg_proc.poll() is None:
            self.ffmpeg_proc.terminate()
        if hasattr(self, 'capture_thread'):
            self.capture_thread.join(timeout=1)
        if hasattr(self, 'process_thread'):
            self.process_thread.join(timeout=1)

    def stop(self):
        """stop system"""
        self.running = False
        self._cleanup()

if __name__ == "__main__":
    system = OddVoiceAssistant()
    
    try:
        if system.start():
            print("system started, listening...")
            while True:
                time.sleep(1)
    except KeyboardInterrupt:
        print("\nstopping system...")
    finally:
        system.stop()
        print("system stopped")
    


from funasr import AutoModel

chunk_size = [0, 10, 5] #[0, 10, 5] 600ms, [0, 8, 4] 480ms
encoder_chunk_look_back = 4 #number of chunks to lookback for encoder self-attention
decoder_chunk_look_back = 1 #number of encoder chunks to lookback for decoder cross-attention

model = AutoModel(model="paraformer-zh-streaming")

import soundfile
import os

wav_file = os.path.join(model.model_path, "example/asr_example.wav")
speech, sample_rate = soundfile.read(wav_file)
chunk_stride = chunk_size[1] * 960 # 600ms


cache = {}
total_chunk_num = int(len((speech)-1)/chunk_stride+1)

print(f"total chunk num={total_chunk_num}, chunk stride={chunk_stride}, len={len(speech)}, type={type(speech)},  dtype={speech.dtype}")

for i in range(total_chunk_num):
    speech_chunk = speech[i*chunk_stride:(i+1)*chunk_stride]
    is_final = i == total_chunk_num - 1

    print(f"chunk {i}, is_final={is_final}, speech_chunk len={len(speech_chunk)}, speech_chunk shape={speech_chunk.shape}, speech_chunk dtype={speech_chunk.dtype}")

    res = model.generate(input=speech_chunk, cache=cache, is_final=is_final, chunk_size=chunk_size, encoder_chunk_look_back=encoder_chunk_look_back, decoder_chunk_look_back=decoder_chunk_look_back)
    print(res)


# -*- coding: utf-8 -*-
""" 
@author: catherine wei
@contact: EMAIL@contact: catherine@oddmeta.com
@software: PyCharm 
@file: result.py 
@time: 2025/6/9 15:00
@info: OddAs websocket protocol define.
"""


class TOddAsrHeader:
    '''
    OddAsr header define.
    ---
    namespace                           String 消息所属的命名空间
    name                                String 消息名称，SentenceBegin 表示一个句子的开始
    status                              Integer 状态码，表示请求是否成功，见服务状态码
    status_text                         String 状态消息
    task_id                             String 任务全局唯一 ID，请记录该值，便于排查问题
    message_id                          String 本次消息的 ID
    '''
    task_id = ""
    message_id = ""

    namespace = ""
    name = ""

    status = 20000000
    # status_message = "" 
    status_text = "Success"             # "GATEWAY|SUCCESS|Success."

    def __init__(self, namespace: str = "",
                 name: str = "",
                 status: int = 0,
                 status_text: str = "",
                 task_id: str = "",
                 message_id: str = "") -> None:
        self.namespace = namespace
        self.name = name
        self.status = status
        self.status_text = status_text
        self.task_id = task_id
        self.message_id = message_id

class TOddAsrPayload:
    '''
    OddAsr transcription result payload
    index                               Integer 句子编号，从 1 开始递增
    time                                Integer 当前已处理的音频时长，单位是毫秒
    result                              String 当前句子的识别结果
    confidence                          Double 当前句子识别结果的置信度，取值范围[0.0, 1.0]，值越大表示置信度越高
    gender                              String 开启 enable_gender_detect 后返回，性别及年龄，包括 male 成年男性,female 成年女性,child 儿童,unknown 未知
    gender_score                        Float 开启 enable_gender_detect 后返回， 算法给出的置信度得分，得分越高可靠性越高，调用方可以基于此得分判断是否采信模型结果
    '''
    index = 0
    time = 0
    result = ""
    confidence = 0.0
    gender = ""
    gender_score = 0.0

    def __init__(self , index: int = 0,
                 time: int = 0,
                 result: str = "",
                 confidence: float = 0.0,
                 gender: str = "",
                 gender_score: float = 0.0) -> None:
        self.index = index
        self.time = time
        self.result = result
        self.confidence = confidence
        self.gender = gender
        self.gender_score = gender_score

class TOddAsrPayloadReq:
    '''
    OddAsr Apply transcription request payload
    ---
    参数 类型 是否必需 说明 
    appkey                              Sring   是 业务方或者业务场景的标记，专有云默认为 default
    format                              String  否 音频编码格式，支持的格式：pcm（无压缩的 pcm 文件或 wav 文件）
    sample_rate                         Integer 否 音频采样率，默认是 16000Hz，请确保采样率与模型类型保持一致
    enable_intermediate_result          Boolean 否 是否返回中间识别结果，默认是 False
    enable_punctuation_prediction       Boolean 否 是否在后处理中添加标点，默认是False
    enable_inverse_text_normalization   Boolean 否 是否在后处理中执行 ITN，设置为 true时，中文数字将转为阿拉伯数字输出，默认是 False

    model                               String  否 模型名称
    customization_id                    String  否 定制模型 ID
    vocabulary_id                       String  否 定制泛热词 ID
    class_vocabulary_id                 Map     否 定制类热词 ID，Map 的 Key 是类名，Value 是类热词词表 ID。目前只有law_politics(政法)，customer_service_8k(客服质检)模型支持人名(PERSON)、地名(ADDRESS)类热词
    max_sentence_silence                Integer 否 语音断句检测阈值，静音时长超过该阈值会被认为断句，合法参数范围 200～2000(ms)，默认值 800ms，开启语义断句(enable_semantic_sentence_detection)后，此参数无效
    enable_semantic_sentence_detection  Boolean 否 是否开启语语义断句，可选，默认是False；开启后强制启用加标点功能且不可禁用，默认开 itn 但可通过传参关闭。
    vad_silence_duration                Integer 否 语义断句长静默的静默时长参数，用户触发此参数阈值后会自动断句，开启语义断句后生效，合法参数范围[2000 10000]ms，默认值 2000ms
    enable_words                        Boolean 否 是否开启返回词信息，默认 False 不开启，在 vad 断句时生效
    disfluency                          Boolean 否 顺滑，默认 False 关闭
    enable_ignore_sentence_timeout      Boolean 否 默认 False，开启后忽略实时转写处理过程中的 asr 超时错误
    enable_gender_detect                Boolean 否 默认 False，开启后返回性别、年龄识别结果，需要单独购买相关算法服务后可以开启
    '''
    appkey = "default"
    format = "pcm"
    sample_rate = 16000
    enable_intermediate_result = False
    enable_punctuation_prediction = False
    enable_inverse_text_normalization = False

    model = ""
    customization_id = ""
    vocabulary_id = ""
    class_vocabulary_id = {}
    max_sentence_silence = 800
    enable_semantic_sentence_detection = False
    vad_silence_duration = 2000
    enable_words = False
    disfluency = False
    enable_ignore_sentence_timeout = False
    enable_gender_detect = False

    def __init__(self , appkey: str = None, 
                 format: str = None, 
                 sample_rate: int = 16000, 
                 enable_intermediate_result: bool = False, 
                 enable_punctuation_prediction: bool = False, 
                 enable_inverse_text_normalization: bool = False, 
                 model: str = "", 
                 customization_id: str = "", 
                 vocabulary_id: str = "", 
                 class_vocabulary_id: dict = None, 
                 max_sentence_silence: int = 800, 
                 enable_semantic_sentence_detection: bool = False, 
                 vad_silence_duration: int = 2000, 
                 enable_words: bool = False, 
                 disfluency: bool = False, 
                 enable_ignore_sentence_timeout: bool = False, 
                 enable_gender_detect: bool = False) -> None:
        self.appkey = appkey
        self.format = format
        self.sample_rate = sample_rate
        self.enable_intermediate_result = enable_intermediate_result
        self.enable_punctuation_prediction = enable_punctuation_prediction
        self.enable_inverse_text_normalization = enable_inverse_text_normalization
        self.model = model
        self.customization_id = customization_id
        self.vocabulary_id = vocabulary_id
        self.class_vocabulary_id = class_vocabulary_id
        self.max_sentence_silence = max_sentence_silence
        self.enable_semantic_sentence_detection = enable_semantic_sentence_detection
        self.vad_silence_duration = vad_silence_duration
        self.enable_words = enable_words
        self.disfluency = disfluency
        self.enable_ignore_sentence_timeout = enable_ignore_sentence_timeout
        self.enable_gender_detect = enable_gender_detect


class TOddAsrWord:
    '''
    参数 类型 说明 
    word        String 识别结果
    begin_time  Integer 词的开始时间，单位是毫秒
    end_time    Integer 词的结束时间，单位是毫秒
    confidence  Double 词的置信度，取值范围[0.0, 1.0]，值越大表示置信度越高
    '''
    word = ""
    begin_time = 0
    end_time = 0
    confidence = 1.0


class TOddAsrPayloadRes:
    '''
    参数 类型 说明 
    index       Integer 句子编号，从 1 开始递增
    time        Integer 当前已处理的音频时长，单位是毫秒
    begin_time  Integer 当前句子句首对应的 SentenceBegin 事件的时间，单位是毫秒
    end_time    Integer 当前句子句尾对应的 SentenceBegin 事件的时间，单位是毫秒
    result      String 当前的识别结果
    confidence  Double 当前句子识别结果的置信度，取值范围[0.0, 1.0]，值越大表示置信度越高    
    fin         Integer 句子是否结束，1 表示结束，0 表示未结束
    words       List<TOddAsrWord> 句子的识别结果，每个元素包含词、开始时间、结束时间、置信度等信息
    '''
    index = 0
    time = 0
    begin_time = 0
    end_time = 0
    result = ""
    confidence = 1.0
    fin = 1
    words = []

class TOddAsrApplyReq:
    '''
    OddAsr Apply transcription request
    ---
    参数 类型 是否必需 说明 
    header                              TOddAsrHeader 是 消息头
    payload                             TOddAsrPayloadReq 是 消息体
    '''
    def __init__(self , header: TOddAsrHeader = None, payload: TOddAsrPayloadReq = None) -> None:
        if header:
            self.header = header
        else:
            self.header = TOddAsrHeader(name="StartTranscription", namespace="SpeechTranscriber", )
        if payload:
            self.payload = payload
        else:
            self.payload = TOddAsrPayloadReq()

class TOddAsrApplyRes:
    def __init__(self , header: TOddAsrHeader = None, payload: TOddAsrPayloadRes = None) -> None:
        if header:
            self.header = header
        else:
            self.header = TOddAsrHeader(name="SentenceBegin", namespace="SpeechTranscriber", )

        if payload:
            self.payload = payload
        else:
            self.payload = TOddAsrPayloadRes()

class TOddAsrTranscribeRes:
    def __init__(self, header: TOddAsrHeader = None, payload: TOddAsrPayloadRes = None) -> None:
        if header:
            self.header = header
        else:
            self.header = TOddAsrHeader(name="TranscriptionResultChanged", namespace="SpeechTranscriber", )
        
        if payload:
            self.payload = payload
        else:
            self.payload = TOddAsrPayloadRes()


# class TCMDCommonRes:
#     name=""
#     message_id=""
#     error_code = 0
#     error_desc=""

#     def __init__(self , message_id, name) -> None:
#         self.message_id = message_id
#         self.name = name

# class TASRRecogTextParam:
#     text=""
#     bg=0
#     ed=0
#     fin=0

#     def __init__(self) -> None:
#         pass

def obj_to_dict(self):
    return dict((name, getattr(self, name)) for name in dir(self) if not name.startswith('__'))

def obj_to_dict_recursive(self):    
    res = {}
    for name in dir(self):
        if not name.startswith('__'):
            attr = getattr(self, name)
            if isinstance(attr, object) and hasattr(attr, '__dict__'):
                res[name] = obj_to_dict_recursive(attr)
            else:
                res[name] = attr
    return res

def obj_from_dict_recursive(self, d):
    for name, value in d.items():
        if isinstance(value, dict):
            setattr(self, name, obj_from_dict_recursive(self, value))
        else:
            setattr(self, name, value)
    return self

if __name__ == "__main__":
    header: TOddAsrHeader = TOddAsrHeader()
    payload: TOddAsrPayloadReq = TOddAsrPayloadReq()
    res = TOddAsrApplyReq(header=header, payload=payload)

    print(f"res={obj_to_dict_recursive(res)}")




#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
# Copyright FunASR (https://github.com/alibaba-damo-academy/FunClip). All Rights Reserved.
#  MIT License  (https://opensource.org/licenses/MIT)

import os
import re
import numpy as np  

PUNC_LIST = ['，', '。', '！', '？', '、', ',', '.', '?', '!']

def pre_proc(text):
    res = ''
    for i in range(len(text)):
        if text[i] in PUNC_LIST:
            continue
        if '\u4e00' <= text[i] <= '\u9fff':
            if len(res) and res[-1] != " ":
                res += ' ' + text[i]+' '
            else:
                res += text[i]+' '
        else:
            res += text[i]
    if res[-1] == ' ':
        res = res[:-1]
    return res

def proc(raw_text, timestamp, dest_text, lang='zh'):
    # simple matching
    ld = len(dest_text.split())
    mi, ts = [], []
    offset = 0
    while True:
        fi = raw_text.find(dest_text, offset, len(raw_text))
        ti = raw_text[:fi].count(' ')
        if fi == -1:
            break
        offset = fi + ld
        mi.append(fi)
        ts.append([timestamp[ti][0]*16, timestamp[ti+ld-1][1]*16])
    return ts
            

def proc_spk(dest_spk, sd_sentences):
    ts = []
    for d in sd_sentences:
        d_start = d['timestamp'][0][0]
        d_end = d['timestamp'][-1][1]
        spkid=dest_spk[3:]
        if str(d['spk']) == spkid and d_end-d_start>999:
            ts.append([d_start*16, d_end*16])
    return ts

def generate_vad_data(data, sd_sentences, sr=16000):
    assert len(data.shape) == 1
    vad_data = []
    for d in sd_sentences:
        d_start = round(d['ts_list'][0][0]/1000, 2)
        d_end = round(d['ts_list'][-1][1]/1000, 2)
        vad_data.append([d_start, d_end, data[int(d_start * sr):int(d_end * sr)]])
    return vad_data

def write_state(output_dir, state):
    for key in ['/recog_res_raw', '/timestamp', '/sentences']:#, '/sd_sentences']:
        with open(output_dir+key, 'w') as fout:
            fout.write(str(state[key[1:]]))
    if 'sd_sentences' in state:
        with open(output_dir+'/sd_sentences', 'w') as fout:
            fout.write(str(state['sd_sentences']))

def load_state(output_dir):
    state = {}
    with open(output_dir+'/recog_res_raw') as fin:
        line = fin.read()
        state['recog_res_raw'] = line
    with open(output_dir+'/timestamp') as fin:
        line = fin.read()
        state['timestamp'] = eval(line)
    with open(output_dir+'/sentences') as fin:
        line = fin.read()
        state['sentences'] = eval(line)
    if os.path.exists(output_dir+'/sd_sentences'):
        with open(output_dir+'/sd_sentences') as fin:
            line = fin.read()
            state['sd_sentences'] = eval(line)
    return state

def convert_pcm_to_float(data):
    if data.dtype == np.float64:
        return data
    elif data.dtype == np.float32:
        return data.astype(np.float64)
    elif data.dtype == np.int16:
        bit_depth = 16
    elif data.dtype == np.int32:
        bit_depth = 32
    elif data.dtype == np.int8:
        bit_depth = 8
    else:
        raise ValueError("Unsupported audio data type")

    # Now handle the integer types
    max_int_value = float(2 ** (bit_depth - 1))
    if bit_depth == 8:
        data = data - 128
    return (data.astype(np.float64) / max_int_value)

def convert_time_to_millis(time_str):
    # 格式: [小时:分钟:秒,毫秒]
    hours, minutes, seconds, milliseconds = map(int, re.split('[:,]', time_str))
    return (hours * 3600 + minutes * 60 + seconds) * 1000 + milliseconds

def extract_timestamps(input_text):
    # 使用正则表达式查找所有时间戳
    timestamps = re.findall(r'\[(\d{2}:\d{2}:\d{2},\d{2,3})\s*-\s*(\d{2}:\d{2}:\d{2},\d{2,3})\]', input_text)
    times_list = []
    print(timestamps)
    # 循环遍历找到的所有时间戳，并转换为毫秒
    for start_time, end_time in timestamps:
        start_millis = convert_time_to_millis(start_time)
        end_millis = convert_time_to_millis(end_time)
        times_list.append([start_millis, end_millis])
    
    return times_list

def convert_time_to_srt_format(time_in_milliseconds):
    hours = time_in_milliseconds // 3600000
    time_in_milliseconds %= 3600000
    minutes = time_in_milliseconds // 60000
    time_in_milliseconds %= 60000
    seconds = time_in_milliseconds // 1000
    time_in_milliseconds %= 1000

    return f"{hours:02}:{minutes:02}:{seconds:02},{time_in_milliseconds:03}"

def text_to_srt(idx, speaker_id, msg, start_microseconds, end_microseconds) -> str:
    start_time = convert_time_to_srt_format(start_microseconds)
    end_time = convert_time_to_srt_format(end_microseconds)

    msg = f"发言人 {speaker_id}: {msg}"
    srt = """%d %s --> %s %s """ % (
        idx,
        start_time,
        end_time,
        msg,
    )
    return srt


if __name__ == '__main__':

    # 示例输入文本
    text = [{'text': '是开始这个呃实时的一个转写。', 'start': 1010, 'end': 4865, 'timestamp': [[1010, 1250], [1750, 1990], [1990, 2230], [2410, 2510], [2510, 2750], [3450, 3690], [3830, 3990], [3990, 4170], [4170, 4250], [4250, 4330], [4330, 4430], [4430, 4570], [4570, 4865]], 'raw_text': '是开始这个呃实时的一个转写', 'spk': 0}, {'text': '对，', 'start': 6040, 'end': 6280, 'timestamp': [[6040, 6280]], 'raw_text': '对', 'spk': 0}, {'text': '然后是转写的一个效果，', 'start': 6640, 'end': 8660, 'timestamp': [[6640, 6800], [6800, 7040], [7360, 7600], [7660, 7820], [7820, 8000], [8000, 8080], [8080, 8160], [8160, 8260], [8260, 8420], [8420, 8660]], 'raw_text': '然后是转写的一个效果', 'spk': 0}, {'text': '大概大概就是这个样子。', 'start': 8680, 'end': 10280, 'timestamp': [[8680, 8760], [8760, 9000], [9160, 9300], [9300, 9480], [9480, 9600], [9600, 9740], [9740, 9840], [9840, 9960], [9960, 10080], [10080, 10280]], 'raw_text': '大概大概就是这个样子', 'spk': 0}, {'text': '然后的话那个在这里边你也可以去给他那个加一个人。', 'start': 10280, 'end': 14500, 'timestamp': [[10280, 10380], [10380, 10500], [10500, 10600], [10600, 10800], [10800, 10900], [10900, 11140], [11180, 11380], [11380, 11500], [11500, 11620], [11620, 11860], [12260, 12400], [12400, 12580], [12580, 12760], [12760, 12980], [12980, 13180], [13180, 13280], [13280, 13480], [13480, 13580], [13580, 13820], [13900, 14080], [14080, 14180], [14180, 14260], [14260, 14500]], 'raw_text': '然后的话那个在这里边你也可以去给他那个加一个人', 'spk': 0}]
    subtitles = []
    for idx, sentence in enumerate(text):
        sub = text_to_srt(idx=idx, speaker_id=sentence['spk'], msg=sentence['text'], start_microseconds=sentence['start'], end_microseconds=sentence['end'])
        subtitles.append(sub)
    print("\n".join(subtitles))

    text = ("0 00:00:01,010 --> 00:00:04,865 发言人 0: 是开始这个呃实时的一个转写。"
            "1 00:00:06,040 --> 00:00:06,280 发言人 0: 对，"
            "2 00:00:06,640 --> 00:00:08,660 发言人 0: 然后是转写的一个效果，"
            "3 00:00:08,680 --> 00:00:10,280 发言人 0: 大概大概就是这个样子。"
            "4 00:00:10,280 --> 00:00:14,500 发言人 0: 然后的话那个在这里边你也可以去给他那个加一个人。"
            "5 00:00:14,660 --> 00:00:19,665 发言人 0: 比如说是嗯我随便给他取一个名字，"
            "6 00:00:20,440 --> 00:00:23,200 发言人 1: 就是连云端的还是自己算的连云端的吧。"
            "7 00:00:23,240 --> 00:00:25,340 发言人 0: 不用连看能调吧。"
            "8 00:00:29,120 --> 00:00:31,480 发言人 2: 这个还有对呀，"
            "9 00:00:32,130 --> 00:00:33,885 发言人 0: 然后这里边可以给他加格。")

    print(extract_timestamps(text))


# -*- coding: utf-8 -*-
""" 
@author: catherine@oddmeta.com 
@software: PyCharm 
@file: sensitivewords.py 
@time: 2021/12/1 16:25 
@info: 敏感词相关处理逻辑
"""
from json import dumps

from oddasr.log import logger
from oddasr.model import db
from oddasr.model import sensitiveword as sw

class SensitiveWordManage(object):
    """
    敏感词管理类
    """

    @classmethod
    def set_sensitive_word(cls, unique_id, sensitive_word):
        """
        设置敏感词
        :param unique_id: 唯一标识
        :param sensitive_word: 敏感词列表
        :return:
        """
        db_session = db.Session()

        sensitive_word_m = db_session.query(sw.CSensitiveWord).filter(sw.CSensitiveWord.unique_id == unique_id).first()
        if sensitive_word_m:
            sensitive_word_m.sensitive_word = dumps(sensitive_word, ensure_ascii=False)
        else:        
            sensitive_word_m = sw.CSensitiveWord(
                unique_id=unique_id,
                sensitive_word=dumps(sensitive_word, ensure_ascii=False)
            )
        db_session.merge(sensitive_word_m)
        db_session.commit()
        data = db.to_dict(sensitive_word_m)
        logger.info(f"set_sensitive_word:{data}")
        db_session.close()
        return data

    @classmethod
    def get_sensitive_word(cls, unique_id):
        """
        获取敏感词
        :param unique_id: 唯一标识
        :return:
        """
        logger.info(f"get_sensitive_word{unique_id}")
        db_session = db.Session()
        sensitive_word_m = db_session.query(sw.CSensitiveWord).filter(sw.CSensitiveWord.unique_id == unique_id).first()
        if not sensitive_word_m:
            db_session.close()
            return None
        data = db.to_dict(sensitive_word_m)
        db_session.close()
        return data

    @classmethod
    def del_sensitive_word(cls, unique_id):
        """
        删除敏感词
        :param unique_id: 唯一标识
        :return:
        """
        db_session = db.Session()
        sensitive_word_m = db_session.query(sw.CSensitiveWord).filter(sw.CSensitiveWord.unique_id == unique_id).first()
        if not sensitive_word_m:
            return
        db_session.delete(sensitive_word_m)
        db_session.commit()
        db_session.close()
        return {}




import mindspore
from mindspore import nn
from mindspore.dataset import vision, transforms
from mindspore.dataset import MnistDataset


# 下载数据集
from download import download

url = "https://mindspore-website.obs.cn-north-4.myhuaweicloud.com/" \
      "notebook/datasets/MNIST_Data.zip"
path = download(url, "./", kind="zip", replace=True)

# 获取数据集对象
train_dataset = MnistDataset('MNIST_Data/train')
test_dataset = MnistDataset('MNIST_Data/test')

# 打印数据集中包含的数据列名，用于dataset的预处理。
print(train_dataset.get_col_names())

# ['image', 'label']

# MindSpore的dataset使用数据处理流水线（Data Processing Pipeline），需指定map、batch、shuffle等操作。这里我们使用map对图像数据及标签进行变换处理，将输入的图像缩放为1/255，根据均值0.1307和标准差值0.3081进行归一化处理，然后将处理好的数据集打包为大小为64的batch。

def datapipe(dataset, batch_size):
    image_transforms = [
        vision.Rescale(1.0 / 255.0, 0),
        vision.Normalize(mean=(0.1307,), std=(0.3081,)),
        vision.HWC2CHW()
    ]
    label_transform = transforms.TypeCast(mindspore.int32)

    dataset = dataset.map(image_transforms, 'image')
    dataset = dataset.map(label_transform, 'label')
    dataset = dataset.batch(batch_size)
    return dataset

# Map vision transforms and batch dataset
train_dataset = datapipe(train_dataset, 64)
test_dataset = datapipe(test_dataset, 64)

# 可使用create_tuple_iterator 或create_dict_iterator对数据集进行迭代访问，查看数据和标签的shape和datatype。

for image, label in test_dataset.create_tuple_iterator():
    print(f"Shape of image [N, C, H, W]: {image.shape} {image.dtype}")
    print(f"Shape of label: {label.shape} {label.dtype}")
    break

# Shape of image [N, C, H, W]: (64, 1, 28, 28) Float32
# Shape of label: (64,) Int32

for data in test_dataset.create_dict_iterator():
    print(f"Shape of image [N, C, H, W]: {data['image'].shape} {data['image'].dtype}")
    print(f"Shape of label: {data['label'].shape} {data['label'].dtype}")
    break

# Shape of image [N, C, H, W]: (64, 1, 28, 28) Float32
# Shape of label: (64,) Int32


# Define model
class Network(nn.Cell):
    def __init__(self):
        super().__init__()
        self.flatten = nn.Flatten()
        self.dense_relu_sequential = nn.SequentialCell(
            nn.Dense(28*28, 512),
            nn.ReLU(),
            nn.Dense(512, 512),
            nn.ReLU(),
            nn.Dense(512, 10)
        )

    def construct(self, x):
        x = self.flatten(x)
        logits = self.dense_relu_sequential(x)
        return logits

model = Network()
print(model)


model.set_train(False)
for data, label in test_dataset:
    pred = model(data)
    predicted = pred.argmax(1)
    print(f'Predicted: "{predicted[:10]}", Actual: "{label[:10]}"')
    break

# Predicted: "[8 8 9 7 2 5 3 4 3 8]", Actual: "[8 8 9 7 2 5 3 9 3 8]"



# -*- coding: utf-8 -*-
""" 
@author: catherine wei
@contact: EMAIL@contact: catherine@oddmeta.com
@software: PyCharm 
@file: odd_asr_sentence.py 
@info: 语音识别类，用于语音识别句子
"""

import torch
import librosa
import torchaudio
import threading

import os
from funasr import AutoModel

from oddasr.logic.utils_speech import convert_pcm_to_float, convert_time_to_millis, text_to_srt
from oddasr.log import logger
import oddasr.odd_asr_config as config

class OddAsrParamsSentence(object):
    def __init__(self, mode="file", hotwords="", return_raw_text=True, is_final=True, sentence_timestamp=False):
        self._mode = mode  # mode should be a string like 'file','stream', 'pipeline'
        self._hotwords = hotwords  # hotwords should be a string like 'word1 word2'
        self._return_raw_text=return_raw_text #return raw text or not, default is True, if False, return json format result, like: {text: "hello world", timestamp: [[0, 1000], [1000, 2000]]}, 
        self._is_final=is_final  #is_final=True, if False, return partial result, like: {text: "hello world", timestamp: [[0, 1000], [1000, 2000]], is_final: False},
        self._sentence_timestamp=sentence_timestamp  #sentence_timestamp=False, if True, return sentence timestamp, like: {text: "hello world", timestamp: [[0, 1000], [1000, 2000]], is_final: False, sentence_timestamp: [[0, 2000]]},
        self._is_busy=False

class OddAsrSentence:
    """
    语音识别类，用于语音识别文件
    """
    _sentenceParam: OddAsrParamsSentence = None
    _model: AutoModel = None
    _device = None

    def __init__(self, sentenceParam:OddAsrParamsSentence=None):

        if sentenceParam is None:
            self._sentenceParam = OddAsrParamsSentence()
        else:
            self._sentenceParam = sentenceParam

        if config.odd_asr_cfg["enable_gpu"]:
            # auto detect GPU _device
            if torch.cuda.is_available():
                self._device = "cuda:0"
            # elif torch.npu.is_available():
            #     self._device = "npu:0"
            else:
                self._device = "cpu"
        else:
            self._device = "cpu"

        # load model on init due to the model is large, and the model is not loaded on the first call
        if config.odd_asr_cfg["preload_model"]:
            self.load_sentence_model(self._device)

        self.lock = threading.Lock()  # mutex lock for _is_busy

    def load_sentence_model(self, device="cuda:0"):
        # load file model
        if self._model is not None:
            return

        logger.info(f"Loading sentence model with device={device}")

        self._model = AutoModel(
            # model="iic/speech_paraformer_asr_nat-zh-cn-16k-aishell2-vocab5212-pytorch",
            model="iic/speech_seaco_paraformer_large_asr_nat-zh-cn-16k-common-vocab8404-pytorch",
            vad_model='iic/speech_fsmn_vad_zh-cn-16k-common-pytorch', vad_model_revision="v2.0.4",
            # punc_model='ct-punc',
            punc_model='iic/punc_ct-transformer_cn-en-common-vocab471067-large', punc_model_revision="v2.0.4",
            spk_model="cam++",
            # spk_model="iic/speech_campplus_sv_zh-cn_16k-common", spk_model_revision="v2.0.4", # 下载模型失败
            # spk_model="iic/speech_campplus_sv_zh-cn_3dspeaker_16k",
            log_level="error",
            hub="ms",  # hub：表示模型仓库，ms为选择modelscope下载，hf为选择huggingface下载。
            device=device,
            disable_update=True
        )
        logger.info("Model loaded successfully.")

    def transcribe_sentence(self, audio_file, hotwords="", output_format="txt"):
        self.set_busy(True)
        try:
            # check audio file exists
            if not os.path.exists(audio_file):
                logger.error(f"Audio file not found: {audio_file}")
                self.set_busy(False)
                raise FileNotFoundError(f"Audio file not found: {audio_file}")
    
            # Check file size
            file_size = os.path.getsize(audio_file)
            if file_size == 0:
                logger.error(f"Audio file is empty: {audio_file}")
                self.set_busy(False)
                raise ValueError(f"Audio file is empty")
            logger.info(f"Audio file size: {file_size} bytes")
    
            try:
                # load audio file
                logger.info(f"Loading audio file: {audio_file}")
                data, sr = librosa.load(audio_file, sr=None, mono=True)
                logger.info(f"Audio loaded successfully. Sample rate: {sr}, data shape: {data.shape}")
                
                # Check if data is valid
                if len(data) == 0:
                    raise ValueError("Loaded audio data is empty")
                    
                data = convert_pcm_to_float(data)
            except Exception as e:
                logger.error(f"Failed to load audio file: {e}")
                self.set_busy(False)
                raise RuntimeError(f"Failed to load audio file: {str(e)}")
    
            # resample audio to 16kHz if necessary
            if sr != 16000:
                try:
                    logger.info(f"Resampling audio from {sr} to 16000 Hz")
                    data = torch.tensor(data, dtype=torch.float32).unsqueeze(0) # Add batch dimension
                    resampler = torchaudio.transforms.Resample(orig_freq=sr, new_freq=16000)
                    data = resampler(data).squeeze(0).numpy()  # Resample and convert to numpy array
                    logger.info(f"Audio resampled to 16000 Hz. Shape: {data.shape}")
                except Exception as e:
                    logger.error(f"Failed to resample audio: {e}")
                    self.set_busy(False)
                    raise RuntimeError(f"Failed to resample audio: {str(e)}")
    
            logger.info(f"Starting speech recognition with expected output_format={output_format}, hotwords: {hotwords}")
    
            if self._model is None:
                try:
                    self.load_file_model(self._device)
                except Exception as e:
                    logger.error(f"Failed to load ASR model: {e}")
                    self.set_busy(False)
                    raise RuntimeError(f"Failed to load ASR model: {str(e)}")
    
            # start speech recognition with hotwords
            try:
                result = self._model.generate(
                    data, 
                    return_raw_text=True, 
                    is_final=True, 
                    sentence_timestamp=False,
                    hotword=hotwords  # Pass the hotwords as a string to the _model
                )
            except Exception as e:
                logger.error(f"ASR sentence generate error: {e}")
                self.set_busy(False)
                raise RuntimeError(f"ASR sentence generate error: {str(e)}")
    
            output_text = ""
            logger.debug(f"ASR result: {result}")
    
            try:
                if output_format == "raw":
                    output_text = result
                elif output_format == "srt":
                    # Check if sentence_info exists in result
                    if isinstance(result, list) and len(result) > 0 and "sentence_info" in result[0]:
                        sentences = result[0]["sentence_info"]
                        subtitles = []

                        logger.debug(f"sentence_info: {sentences[:2]}...")

                        for idx, sentence in enumerate(sentences):
                            sub = text_to_srt(idx=idx, speaker_id=sentence['spk'], 
                                            msg=sentence['text'], start_microseconds=sentence['start'], 
                                            end_microseconds=sentence['end'])
                            subtitles.append(sub)

                        output_text = "\n".join(subtitles)
                    else:
                        output_text = result[0]["text"] if isinstance(result, list) and len(result) > 0 and "text" in result[0] else ""
                elif output_format == "spk":
                    # Check if sentence_info exists in result
                    if isinstance(result, list) and len(result) > 0 and "sentence_info" in result[0]:
                        sentences = result[0]["sentence_info"]
                        subtitles = []

                        for idx, sentence in enumerate(sentences):
                            sub = f"发言人 {sentence['spk']}: {sentence['text']}"
                            subtitles.append(sub)

                        output_text = "\n".join(subtitles)
                    else:
                        output_text = result[0]["text"] if isinstance(result, list) and len(result) > 0 and "text" in result[0] else ""
                elif output_format == "txt":
                    output_text = result[0]["text"] if isinstance(result, list) and len(result) > 0 and "text" in result[0] else ""
                else:
                    output_text = result[0]["text"] if isinstance(result, list) and len(result) > 0 and "text" in result[0] else ""
            except Exception as e:
                logger.error(f"Failed to process output format: {e}")
                self.set_busy(False)
                raise RuntimeError(f"Failed to process output format: {str(e)}")
                
            self.set_busy(False)
            return output_text
    
        except Exception as e:
            self.set_busy(False)
            logger.error(f"ASR sentence generate error: {e}")
            raise RuntimeError(f"ASR sentence generate error: {e}")

    def set_busy(self, is_busy):
        with self.lock:  # 使用锁保护共享资源
            self._sentenceParam._is_busy = is_busy
            if not is_busy:
                logger.info(f"set_busy to False, done")

    def is_busy(self):
        with self.lock:  # 使用锁保护共享资源
            return self._sentenceParam._is_busy



from funasr import AutoModel
# from loguru import logger

import logging
logging.basicConfig(level=logging.DEBUG)
logging.getLogger("funasr").setLevel(logging.DEBUG)
logging.getLogger("onnxruntime").setLevel(logging.ERROR)
logging.getLogger("onnxruntime.capi.onnxruntime_pybind11_state").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)


class FunASR:
    def __init__(self):
        self.__model = None

    def __init_model(self):
        if self.__model:
            return

        logger.debug("funasr :: init model start")
        self.__model = AutoModel(model="paraformer-zh",
                                 vad_model="fsmn-vad",
                                 punc_model="ct-punc",
                                 spk_model="cam++",
                                 log_level="error",
                                 hub="ms"  # hub：表示模型仓库，ms为选择modelscope下载，hf为选择huggingface下载。
                                 )
        logger.debug("funasr :: init model complete")

    def __convert_time_to_srt_format(self, time_in_milliseconds):
        hours = time_in_milliseconds // 3600000
        time_in_milliseconds %= 3600000
        minutes = time_in_milliseconds // 60000
        time_in_milliseconds %= 60000
        seconds = time_in_milliseconds // 1000
        time_in_milliseconds %= 1000

        return f"{hours:02}:{minutes:02}:{seconds:02},{time_in_milliseconds:03}"

    def __text_to_srt(self, idx, speaker_id, msg, start_microseconds, end_microseconds) -> str:
        start_time = self.__convert_time_to_srt_format(start_microseconds)
        end_time = self.__convert_time_to_srt_format(end_microseconds)

        msg = f"发言人 {speaker_id}: {msg}"
        srt = """%d
%s --> %s
%s
            """ % (
            idx,
            start_time,
            end_time,
            msg,
        )
        return srt

    def transcribe(self, audio_file: str, output_format: str = "txt"):
        self.__init_model()
        logger.info(f"funasr :: start transcribe audio file: {audio_file}")

        res = self.__model.generate(input=audio_file, batch_size_s=300)
        text = res[0]['text']
        logger.info(f"funasr :: complete transcribe audio file: {audio_file}")
        match output_format:
            case "srt":
                sentences = res[0]['sentence_info']
                subtitles = []

                for idx, sentence in enumerate(sentences):
                    sub = self.__text_to_srt(idx, sentence['spk'], sentence['text'], sentence['start'], sentence['end'])
                    subtitles.append(sub)

                return "\n".join(subtitles)
            case "txt":
                speaker_info = [f"发言人 {sentence['spk']}: {sentence['text']}" for sentence in res[0]['sentence_info']]
                return "\n".join(speaker_info)
            case _:
                return text


funasr = FunASR()

text=funasr.transcribe('aketao.wav', output_type="srt")

print(text)




# -*- coding: utf-8 -*-
""" 
@author: catherine wei
@contact: EMAIL@contact: catherine@oddmeta.com
@software: PyCharm 
@file: odd_slp_server.py 
@info: 自学习平台
"""

"""Server example using the asyncio API."""

import asyncio
from websockets.asyncio.server import serve
import json
import websockets
import uuid
import queue

import oddasr.odd_asr_config as config
from oddasr.log import logger

from oddasr.logic.odd_asr_exceptions import mai_err_name, EM_ERR_ASR_ARGS_ERROR
from oddasr.logic.odd_asr_exceptions import *
from oddasr.logic.proto import TOddAsrTranscribeRes, obj_to_dict, TOddAsrApplyRes, obj_from_dict_recursive, obj_to_dict_recursive

'''
client --> server: TCmdApppyAsrReq
server --> client: TCmdApppyAsrRsp with task_id
server --> server: add client to clients
client --> server: PCMData
server --> client: ASRResult

'''


class OddSlpServer:
    def __init__(self):
        self._clients_set = set()
        self._sessionid_set = set()
        self._conn_sessionid = dict()
        self._sessionid_conn = dict()

    async def handle_client(self, *args, **kwargs):
        # 从参数中提取 websocket 和 path
        websocket = args[0]
        logger.debug(f"Client connected: {websocket}, args={args}, len={len(args)}, kwargs={kwargs}")
        
        try:
            async for message in websocket:
                if not websocket in self._clients_set:
                    self._clients_set.add(websocket)
                else:
                    '''
                    客户端已经申请过ASR，并且已经连接上了，此时收到的消息是PCMData
                    '''
                    self.onRecv(websocket, message)
                    continue

        finally:
            self.onClose(websocket)

    async def doSend(self, websocket, message):
        '''
        发送消息给客户端
        :param websocket:
        :param message:
        :return:
        '''
        try:
            # if not isinstance(message, str):
            #     message = json.dumps(message)
            #     logger.debug(f"doSend: {message}")
            # else:
            #     logger.debug(f"doSend: {len(message)}")
            await websocket.send(message)
        except websockets.exceptions.ConnectionClosedOK:
            logger.error(f"Connection closed normally: {websocket}")
        except websockets.exceptions.ConnectionClosedError as e:
            logger.error(f"Connection closed with error: {e}, {websocket}")
        except Exception as e:
            logger.error(f"Unexpected error in doSend: {e}, {websocket}")

    async def doBroadcast(self, message):
        for client in self._clients_set:
            await client.send(message)

    def onRecv(self, websocket, pcm_data):
        logger.debug(f"onRecv: {len(pcm_data)}, websocket={websocket}")

    def onClose(self, websocket):
        logger.info(f"Client disconnected: {websocket}")
        if websocket in self._clients_set:
            self._clients_set.remove(websocket)

            logger.info(f"remove client={websocket}")
        else:
            logger.error(f"client={websocket} not in clients_set")

    def doInit(self, websocket, message):
        # 解析json，若第一个消息不是json，则关闭连接
        result = False
        try:
            req = json.loads(message)
        except Exception as e:
            logger.error(f"Invalid json format. Received message: {message}. webocket={websocket}")
            message_id = ''
            name = ''

            res = TOddAsrApplyRes()
            res.header.message_id = message_id
            res.header.name = name
            res.header.status = EM_ERR_ASR_ARGS_ERROR
            res.header.status_text = mai_err_name(EM_ERR_ASR_ARGS_ERROR)

            return result, res, ""

        # 解析json中的session_id
        res = TOddAsrApplyRes()
        res.header.message_id = req['message_id'] if 'message_id' in req else ''
        res.header.task_id = req['task_id'] if 'task_id' in req else ''
        res.header.name = req['name'] if 'name' in req else ''

        # 若第一个消息不是StartTranscription，则关闭连接
        if req['name'] != 'StartTranscription':
            logger.error(f"Invalid name. Received message: {message}, req['name']")
            res.header.status = EM_ERR_ASR_ARGS_ERROR
            res.header.status_text = mai_err_name(EM_ERR_ASR_ARGS_ERROR)

            return result, res, res.header.task_id

        if "task_id" in req and req['task_id']:
            if req['task_id'] in self._sessionid_set:
                '''若task_id已经存在，说明是之前已经申请过ASR，但是中间网络异常，并断线重连了
                '''
                self._sessionid_conn[req['task_id']] = websocket
                self._conn_sessionid[websocket] = req['task_id']

                res.header.status = 0
                res.header.status_text = "Success"

                result = True
            else:
                '''
                若task_id不存在，说明是一个非法请求
                '''
                logger.error(f"Received message: {message}, req['task_id']={req['task_id']}")
                res.header.status  = EM_ERR_ASR_SESSION_ID_NOVALID
                res.header.status_text = mai_err_name(EM_ERR_ASR_SESSION_ID_NOVALID)
                result = False
        else:
            '''
            若task_id为空，说明是第一次申请ASR，需要生成一个task_id
            '''
            res.header.task_id = str(uuid.uuid1())
            res.header.name = "SentenceBegin"
            res.header.status = 0

            self._sessionid_set.add(res.header.task_id)
            self._sessionid_conn[res.header.task_id] = websocket
            self._conn_sessionid[websocket] = res.header.task_id

            result = True

        return result, res, res.header.task_id


    async def send(self, websocket, message):
        await websocket.send(message)

async def notify_task(_wss_server=None):
    global asr_result_queue
    while True:
        try:
            # 使用 asyncio.to_thread 从队列获取消息
            message = await asyncio.to_thread(asr_result_queue.get, timeout=1)
            logger.debug(f"notifyTask: {message.text}")

            if _wss_server:
                if message.webocket in _wss_server._clients_set:
                    # 发送消息给客户端
                    res = TOddAsrTranscribeRes()
                    res.header.message_id = message.message_id
                    res.header.name = message.name
                    res.header.status = message.status
                    res.header.status_text = message.status_text
                    res.payload = message.payload
                    message.text = json.dumps(obj_to_dict_recursive(res))
                    
                    logger.debug(f"notifyTask: send to client={message.webocket}, message.text={message.text}")
                    await _wss_server.doSend(message.webocket, message.text)
        except queue.Empty:
            # 队列为空，继续等待
            continue
        except Exception as e:
            # 处理其他异常
            logger.error(f"notifyTask error: {e}")
            continue


def init_instances_stream(server: OddSlpServer):
    '''
    初始化odd_asr_stream实例。
    由于初始化加载模型比较耗时，所以在启动的时候就预加载。
    电脑内存太小，默认这里只初始化2个实例。
    每个odd_asr_stream实例对应一个websocket连接。
    每个websocket连接对应一个odd_asr_stream实例。
    每个odd_asr_stream实例对应一个session_id。
    每个session_id对应一个websocket连接。
    每个websocket连接对应一个session_id。
    每个session_id对应一个odd_asr_stream实例。
    :param server:
    :return:
    '''
    pass

def init_notify_task(server: OddSlpServer):
    '''
    初始化notifyTask
    :param server:
    :return:
    '''
    # notify_Task = notifyTask()
    # notify_Task.start(server)
    pass

async def start_slp_server():
    global _wss_server
    _wss_server = OddSlpServer()

    init_notify_task(_wss_server)
    init_instances_stream(_wss_server)

    async with serve(_wss_server.handle_client, config.WS_HOST, config.WS_PORT):
        await asyncio.Future()  # run forever    

if __name__ == "__main__":
    asyncio.run(start_slp_server())



import os
import glob
from Cython.Build import cythonize
from setuptools import Extension
from setuptools import setup
import numpy

import shutil

def copy_file(src, dst):
    """支持文件和目录的复制，如果目标存在则强制覆盖
    Args:
        src: 源文件或目录路径
        dst: 目标文件或目录路径
    """
    try:
        # 检查源是否存在
        if not os.path.exists(src):
            print(f"源路径不存在: {src}")
            return
        
        # 如果源是文件
        if os.path.isfile(src):
            # 确保目标目录存在
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            
            # 强制覆盖复制文件
            shutil.copy2(src, dst)
            print(f'文件复制成功: {src} -> {dst}')
        
        # 如果源是目录
        elif os.path.isdir(src):
            # 如果目标目录已存在，先删除
            if os.path.exists(dst):
                shutil.rmtree(dst)
                print(f'删除已存在的目标目录: {dst}')
            
            # 递归复制目录
            shutil.copytree(src, dst)
            print(f'目录复制成功: {src} -> {dst}')
    
    except Exception as e:
        print(f'复制失败: {e}')


def get_build_target_dir():

    # return os.path.join(os.path.dirname(__file__), "build")
    # 获取build目录下的第一层子目录
    build_dir = "build"
    sub_dirs = [d for d in os.listdir(build_dir) if os.path.isdir(os.path.join(build_dir, d))]

    # 或者使用glob模式匹配（推荐）
    sub_dirs = glob.glob(os.path.join(build_dir, "lib.*"))

    if sub_dirs:
        # 获取第一个匹配的目录名（通常只有一个）
        platform_dir = os.path.basename(sub_dirs[0])
        print(f"获取到的平台目录名: {platform_dir}")
        
        # 完整路径
        full_path = os.path.join(build_dir, platform_dir)
        print(f"完整路径: {full_path}")
    else:
        print("未找到build目录下的子目录")

    return full_path

# 获取所有需要编译的Python文件
SRC_FILES = []

# 查找所有下级目录中的.py文件，不包括当前目录
for py_file in glob.glob("*/**/*.py", recursive=True):
    # 排除不需要编译的文件
    if not py_file.startswith("venv") and \
       not py_file.startswith(".venv") and \
       not py_file.startswith("logs") and \
       not py_file.startswith("build") and \
       not py_file.startswith("charts") and \
       not py_file.startswith("tests"):

        print(f"加入待编译文件: {py_file}")
        SRC_FILES.append(py_file)

# 创建扩展对象列表
extensions = []
for file in SRC_FILES:
    # 创建扩展模块名称（将路径分隔符替换为点）
    module_name = os.path.splitext(file)[0].replace(os.path.sep, '.')
    extensions.append(Extension(module_name, [file]))

# 编译所有扩展模块
if __name__ == '__main__':
    print("开始编译...")

    setup(
        name="oddasr",
        ext_modules=cythonize(
            SRC_FILES, 
            compiler_directives={
                'language_level': 3,
                'boundscheck': False,
                'wraparound': False
            }),
        zip_safe=False,
        include_dirs=[numpy.get_include()]
    )
    print("编译完成！")

    # 将根目录下的所有.py文件复制到build目录下
    print("复制基础文件列表到build目录...")
    for py_file in glob.glob("*.py"):
        build_target_dir = get_build_target_dir()
        if not py_file.startswith("venv") and \
           not py_file.startswith(".venv") and \
           not py_file.startswith("logs") and \
           not py_file.startswith("build") and \
           not py_file.startswith("charts") and \
           not py_file.startswith("tests"):
            print(f"复制文件: {py_file}")
            copy_file(py_file, f"{build_target_dir}/{py_file}")
    
    copy_file("templates", f"{build_target_dir}/templates")
    copy_file("static", f"{build_target_dir}/static")
    copy_file("tests", f"{build_target_dir}/tests")
    copy_file("requirements.txt", f"{build_target_dir}/requirements.txt")
    copy_file("start.bat", f"{build_target_dir}/start.bat")
    copy_file("start.sh", f"{build_target_dir}/start.sh")
    print("复制完成！")



from pydoc import text
from sqlalchemy import Column, Integer, String,Text
# from sqlalchemy.dialects.mysql import MEDIUMTEXT, LONGTEXT

from oddasr.model import db


class CHotWords(db.Base):
    __tablename__ = 'oddasr_hotwords'
    id = Column(Integer, primary_key=True)
    hotwords_id = Column(String(64))    #howords_id:unique_type
    hotwords_type = Column(Integer, default=None)  # 类型 0:人名 1:地名 2:专有名词
    words = Column(Text)

    def __repr__(self):
        return f'<CHotWords {self.hotwords_id!r}>'
